/*
 * linux/driver/ide/arm/ide-armadillo3x0.c
 *
 * Copyright (C) 2006 Atmark Techno, Inc.
 */

#include <linux/config.h>
#include <linux/ide.h>
#include <asm/arch/hardware.h>
#include <asm/arch/armadillo3x0_misc.h>

#define MISCREG_IDEPIO3 (1 << 0)
#define MISCREG(type,off) (*(volatile type *)(0xf6000000 + (off)))
#define MISCREG_IDEMODE   MISCREG(u8, 0x4)

#define DEVICE_NAME "ide-armadillo3x0"

//#define DEBUG
#if defined(DEBUG)
#define DEBUG_FUNC() printk(DEVICE_NAME ": %s()\n", __FUNCTION__)
#else
#define DEBUG_FUNC()
#endif
#define PRINT_INFO(args...) printk(KERN_INFO DEVICE_NAME ": " args)
#define PRINT_ERR(args...) printk(KERN_ERR DEVICE_NAME ": " args)

volatile u8 miscreg_idemode = 0x0;

static int miscreg_pio = -1;

static void armadillo3x0_ide_tuneproc (ide_drive_t *drive, byte pio)
{
  int pio_mode;

  DEBUG_FUNC();
  
  if (drive->id) {
    //if ((pio_mode = ide_scan_pio_blacklist(drive->id->model)) == -1) {
    pio_mode = ide_get_best_pio_mode(drive, pio, 4, NULL);
    if (pio_mode == -1) {
      if ((drive->id->field_valid & 2) && (drive->id->capability & 8) &&
	  (drive->id->eide_pio_modes & 7)) {
	    pio_mode = 3;
      } else {
	pio_mode = 0;
      }
    } else {
      pio_mode = (pio_mode < 3) ? 0 : 3;
    }
  } else {
    pio_mode = 0;
  }

  if (pio_mode >= 3) {
    if (ide_config_drive_speed (drive, XFER_PIO_3)) {
      PRINT_ERR("ide_config_drive_speed error.\n");
      pio_mode = 0;
    }
  }
  if (pio_mode < 3) {
    if (miscreg_pio != 0) {
      miscreg_pio = 0;
      miscreg_idemode &= ~MISCREG_IDEPIO3;
      MISCREG_IDEMODE = miscreg_idemode;
      PRINT_INFO("Set PIO mode 0.\n");
    }
  } else {
    if (miscreg_pio < 0) {
      miscreg_pio = 3;
      miscreg_idemode |= MISCREG_IDEPIO3;
      MISCREG_IDEMODE = miscreg_idemode;
      PRINT_INFO("Set PIO mode 3.\n");
    }
  }
}

void armadillo3x0_ide_init_hwifs(void)
{
  int i, j;

  DEBUG_FUNC();
  
  for (i = 0; i < MAX_HWIFS; i++) {
    for (j = 0; j < MAX_DRIVES; j++) {
      ide_hwifs[i].drives[j].autotune = IDE_TUNE_AUTO;
    }
    ide_hwifs[i].tuneproc = armadillo3x0_ide_tuneproc;
  }
}

static int
armadillo3x0_ide_probe(void)
{
	unsigned char state = readb(A3X0_MISC_IDEMCR);
	return ((state & 0x02) ? 0 : -1);
}

int armadillo3x0_ide_init_hwif_ports(hw_regs_t *hw,
				      unsigned long io_addr,
				      unsigned long ctl_addr)
{
  unsigned long reg = io_addr;
  int         i;

  if (armadillo3x0_ide_probe())
	  return -1;
  
  memset (hw, 0, sizeof(*hw));
  
  for (i = IDE_DATA_OFFSET; i <= IDE_STATUS_OFFSET; i++) {
    hw->io_ports[i] = reg;
    reg += 2;
  }
  if (ctl_addr) {
    hw->io_ports[IDE_CONTROL_OFFSET] = ctl_addr;
  } else {
    hw->io_ports[IDE_CONTROL_OFFSET] = io_addr + 0x2c;
  }

  hw->io_ports[IDE_IRQ_OFFSET] = 0;

  return 0;
}
